from functools import update_wrapper
from typing import Any, Callable, Optional


class classproperty:
    """
    A decorator that converts a method to a `classproperty`

    Example Usage:
    --------------

    >>> from random import random
    >>> class Foo:
    ...     @classproperty
    ...     def bar(cls):
    ...         return
    ...
    >>> Foo.bar
    0.6597781045639168
    >>> Foo.bar
    0.45628230476739395
    """

    def __init__(self, method: Callable[..., Any]) -> None:
        """Create a `classproperty` instance

        Args:
            method (Callable[..., Any]): the method being decorated
        """
        self.method = method

        update_wrapper(self, method)

    def __get__(self, obj: Optional[Any], objtype: Optional[Any] = None) -> Any:
        return self.method(objtype)


class cached_classproperty(classproperty):
    """
    A decorator that converts a method to a `cached_classproperty`.
    As indicated by the name, once the method is called, its
    return value is cached and is therefore returned unchanged.

    Example Usage:
    --------------

    >>> from random import random
    >>> class Foo:
    ...     @cached_classproperty
    ...     def bar(cls):
    ...         return
    ...
    >>> Foo.bar
    0.6597781045639168
    >>> Foo.bar
    0.6597781045639168
    """

    def __init__(self, method: Callable[..., Any]) -> None:
        """Create a `cached_classproperty` instance

        Args:
            method (Callable[..., Any]): the method being decorated
        """
        super().__init__(method)

        self.cache = None

    def __get__(self, obj: Optional[Any], objtype: Optional[Any] = None) -> Any:
        if self.cache is None:
            self.cache = super().__get__(obj, objtype=objtype)

        return self.cache



import logging
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Iterator

import dotify.models as models
from dotify.model import Model, logger

logger = logging.getLogger(f"{logger.name}.{__name__}")

if TYPE_CHECKING is True:
    from dotify.models.track import Track


class Playlist(Model):
    """ """

    class Json:
        """ """

        dependencies = ["dotify.models.User", "dotify.models.Image"]

    def __init__(self, **props) -> None:
        if "tracks" in props:
            del props["tracks"]

        super().__init__(**props)

    def __str__(self):
        return self.name

    def __iter__(self):
        return self.tracks

    @property
    def url(self):
        """ """
        return self.external_urls.spotify

    @property
    def tracks(self) -> Iterator["Track"]:
        """ """
        response, offset = (
            self.context.playlist_items(self.url, additional_types=("track",)),
            0,
        )

        while True:
            for result in response["items"]:
                url = result["track"]["external_urls"]["spotify"]

                yield models.Track.from_url(url)

            offset += len(response["items"])

            if response["next"] is None:
                break

            response = self.context.playlist_items(
                self.url, additional_types=("track",), offset=offset
            )

    def download(
        self, path: PathLike, skip_existing: bool = False, logger: None = None
    ) -> PathLike:
        """"""
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)

        for track in self.tracks:
            track.download(
                path / f"{track}.mp3", skip_existing=skip_existing, logger=logger
            )

        return path

    @classmethod
    @Model.validate_url
    @Model.http_safeguard
    def from_url(cls, url: str) -> "Playlist":
        """"""
        return cls(**cls.context.playlist(url))



import logging
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator

import requests
from mutagen.id3 import APIC

import dotify.models as models
from dotify.model import Model, logger

logger = logging.getLogger(f"{logger.name}.{__name__}")

if TYPE_CHECKING is True:
    from dotify.models.artist import Artist
    from dotify.models.track import Track


class Album(Model):
    """ """

    class Json:
        """ """

        dependencies = [
            "dotify.models.Track",
            "dotify.models.Artist",
            "dotify.models.Image",
        ]

    def __str__(self):
        return f"{self.artist} - {self.name}"

    def __iter__(self):
        return self.tracks

    @property
    def artist(self) -> "Artist":
        """ """
        return self.artists[0]

    @property
    def url(self):
        """ """
        return self.external_urls.spotify

    @property
    def cover(self) -> Any:
        """ """
        response = requests.get(self.images[0].url)

        assert response.status_code == 200, f"Failed to fetch {self.images[0].url}"

        return APIC(
            encoding=3, mime="image/jpeg", type=3, desc="Cover", data=response.content
        )

    @property
    def tracks(self) -> Iterator["Track"]:
        """ """
        response, offset = self.context.album_tracks(self.url), 0

        while True:
            for result in response["items"]:
                url = result["external_urls"]["spotify"]

                yield models.Track.from_url(url)

            offset += len(response["items"])

            if response["next"] is None:
                break

            response = self.context.album_tracks(self.url, offset=offset)

    def download(
        self, path: PathLike, skip_existing: bool = False, logger: None = None
    ) -> PathLike:
        """"""
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)

        for track in self.tracks:
            track.download(
                path / f"{track}.mp3", skip_existing=skip_existing, logger=logger
            )

        return path

    @classmethod
    @Model.validate_url
    @Model.http_safeguard
    def from_url(cls, url: str) -> "Album":
        """"""
        return cls(**cls.context.album(url))



import logging

from dotify.model import Model, logger

logger = logging.getLogger(f"{logger.name}.{__name__}")


class Image(Model):
    """ """

    class Json:
        """ """

    def __str__(self):
        return self.url



import logging
import threading
from typing import Any, Dict, List

from spotipy import Spotify as Client
from spotipy.client import logger
from spotipy.oauth2 import SpotifyClientCredentials

logger = logging.getLogger(f"{logger.name}.{__name__}")


class Dotify(Client):
    """
    Example Usage:
        >>> from dotify import Dotify
        ... with Dotify(spotify_client, spotify_secret):
        ...     for result in Track.search(query):
        ...         ...
    """

    _context = threading.local()

    def __init__(self, client_id: str, client_secret: str) -> None:
        """Create a `Dotify` instance

        Args:
            client_id (str): your Spotify API client ID
            client_secret (str): your Spotify API client secret
        """
        super().__init__(
            client_credentials_manager=SpotifyClientCredentials(
                client_id=client_id, client_secret=client_secret
            )
        )

    def __del__(self):
        if hasattr(self, "_session"):
            super().__del__()

    def __enter__(self) -> "Dotify":
        type(self).get_contexts().append(self)

        return self

    def __exit__(self, exc_type: None, exc_value: None, exc_trace: None) -> None:
        type(self).get_contexts().pop()

        if exc_type is not None:
            logger.error("%s: %s", exc_type.__name__, exc_value)

    @classmethod
    def get_contexts(cls) -> List["Dotify"]:
        """Get the `Dotify` context stack

        Returns:
            List[Dotify]: the `Dotify` context stack
        """
        if not hasattr(cls._context, "stack"):
            cls._context.stack = []

        return cls._context.stack

    @classmethod
    def get_context(cls) -> "Dotify":
        """Get the topmost context from the stack

        Raises:
            TypeError: if the context stack is empty

        Returns:
            Dotify: the topmost context
        """
        try:
            return cls.get_contexts()[-1]
        except IndexError:
            raise TypeError("No context on context stack")

    def search(self, type: str, query: str, limit: int = 1) -> List[Dict[str, Any]]:
        """Perform a Spotify search given a `query`

        Args:
            type (str): One of 'artist', 'album', 'track', 'playlist'
            query (str): the search `query`
            limit (int, optional): the number of items to return. Defaults to 1.

        Returns:
            List[Dict[str, Any]]: A list containing the search results
        """
        results = super().search(query, type=type, limit=limit)

        return results[f"{type}s"]["items"]



import json
import logging
from abc import ABCMeta
from os import PathLike
from typing import Any, Dict, Optional

from python_jsonschema_objects import ObjectBuilder
from python_jsonschema_objects.classbuilder import LiteralValue, ProtocolBase
from python_jsonschema_objects.validators import ValidationError
from python_jsonschema_objects.wrapper_types import ArrayWrapper

logger = logging.getLogger(__name__)


class JsonSerializableMeta(ABCMeta):
    """
    A metaclass responsible for resolving a class' JSON schema
    and defining the class at hand based on it
    """

    def __new__(cls, name, bases, attrs):
        if "Json" not in attrs or not hasattr(attrs["Json"], "schema"):
            return super().__new__(cls, name, bases, attrs)

        path = attrs["Json"].schema.absolute()
        with path.open() as file:
            json_schema = json.load(file)

            builder = ObjectBuilder(str(path))

            classes = builder.build_classes(
                strict=True, named_only=True, standardize_names=False
            )

            json_schema = getattr(classes, name)

        return super().__new__(cls, name, (*bases, json_schema), attrs)


class JsonSerializable(ProtocolBase, metaclass=JsonSerializableMeta):
    """
    A class providing JSON serialization and de-serialization
    """

    class Json:
        schema: PathLike
        dependencies: Dict[str, "JsonSerializable"]

    def __setattr__(self, name: str, val: Any) -> None:
        try:
            super().__setattr__(name, val)
        except ValidationError:
            if hasattr(self, "__annotations__") and name in self.__annotations__:
                self.__dict__[name] = val
            else:
                raise

    @classmethod
    def _resolve_dependency(cls, obj: Any) -> Optional["JsonSerializable"]:
        """Given a `python_jsonschema_objects` it returns
        the `JsonSerializable` dependency that has been registered
        with it, given that it's available

        Returns:
            Optional["JsonSerializable"]: the corresponding `JsonSerializable`
            dependency
        """
        if not hasattr(cls.Json, "dependencies"):
            return None

        return cls.Json.dependencies.get(obj.__class__.__name__, None)

    def __getattribute__(self, name: str) -> Any:
        obj = super().__getattribute__(name)

        if isinstance(obj, LiteralValue):
            return obj._value
        if isinstance(obj, ArrayWrapper):
            array = []
            for element in obj.typed_elems:
                dependency = self._resolve_dependency(element)
                if dependency is None:
                    array.append(element)
                else:
                    array.append(dependency(**element.as_dict()))

            return array
        if isinstance(obj, ProtocolBase):
            dependency = self._resolve_dependency(obj)

            if dependency is not None:
                return dependency(**obj.as_dict())

            return obj

        return obj

    def __getattr__(self, name: str):
        if name in self.__prop_names__:
            return self._properties[name]
        if name in self._extended_properties:
            return self._extended_properties[name]

        return object.__getattribute__(self, name)



import logging

from dotify.model import Model, logger

logger = logging.getLogger(f"{logger.name}.{__name__}")


class User(Model):
    """A model representing a Spotify User"""

    class Json:
        pass

    def __str__(self):
        return self.display_name



from dotify.dotify import Dotify
from dotify.models.album import Album
from dotify.models.playlist import Playlist
from dotify.models.track import Track

__all__ = [
    "Dotify",
    "Album",
    "Playlist",
    "Track",
]



import logging

from dotify.model import Model, logger

logger = logging.getLogger(f"{logger.name}.{__name__}")


class Artist(Model):
    """ """

    class Json:
        """ """

    @property
    def url(self):
        """ """
        return self.external_urls.spotify

    def __str__(self) -> str:
        return self.name



import logging
from functools import wraps
from importlib import import_module
from os import PathLike
from pathlib import Path
from re import match
from typing import Any, Callable, Iterator

from spotipy.exceptions import SpotifyException

from dotify.decorators import cached_classproperty
from dotify.dotify import Dotify
from dotify.json_serializable import JsonSerializable, JsonSerializableMeta, logger

logger = logging.getLogger(f"{logger.name}.{__name__}")


class ModelMeta(JsonSerializableMeta):
    """
    A metaclass serving as an abstraction layer over
    the `JsonSerializableMeta` metaclass, that automatically
    resolves the path to the `Model`'s JSON schema, as well as
    builds the `dependency` dictionary
    """

    @classmethod
    def dependency_basename(cls, model_name: str) -> str:
        """Given the name of a `Model` resolve the basename of
        the corresponding json schema

        Args:
            model_name (str): the name of a `Model`

        Returns:
            str: the basename of the file containing the json schema
        """
        return f"{model_name.lower()}.json"

    @classmethod
    def dependency_path(cls, model_name: str) -> PathLike:
        """Given the name of a `Model` resolve the path to
        the corresponding json schema

        Args:
            model_name (str): the name of a `Model`

        Returns:
            PathLike: the path to the file containing the json schema
        """
        return (
            Path(__file__).parent
            / "models"
            / "schema"
            / cls.dependency_basename(model_name)
        )

    def __new__(cls, name, bases, attrs):
        if "Json" in attrs:
            attrs["Json"].schema = cls.dependency_path(name)

            if hasattr(attrs["Json"], "dependencies"):
                dependency_names = attrs["Json"].dependencies

                @cached_classproperty
                def dependencies(_):
                    types = []
                    for dependency in dependency_names:
                        module, _, type = dependency.rpartition(".")

                        module = import_module(module)
                        type = getattr(module, type)

                        types.append(type)

                    return {
                        cls.dependency_basename(dependency.__name__): dependency
                        for dependency in types
                    }

                attrs["Json"].dependencies = dependencies

        return super().__new__(cls, name, bases, attrs)


class Model(JsonSerializable, metaclass=ModelMeta):
    """
    The base class for every Spotify Web API entity
    """

    class UnexpectedError(Exception):
        """An exception indicating an unexpected error"""

        pass

    class InvalidURL(Exception):
        """
        An exception thrown if the provided URL does not
        correspond to a valid Spotify URL
        """

        pass

    class NotFound(Exception):
        """
        An exception thrown if an operation fails to
        retrieve the necessary information
        """

        pass

    def __repr__(self):
        return f'<{self.__class__.__name__} "{str(self)}">'

    @cached_classproperty
    def context(cls) -> Dotify:
        """Get the current `Dotify` context

        Returns:
            Dotify: the current `Dotify` context
        """
        return Dotify.get_context()

    @classmethod
    def view_name(cls) -> str:
        """Return the name of the Spotify view corresponding to the `Model`

        Returns:
            str: the name of the Spotify view
        """
        return cls.__name__.lower()

    @classmethod
    def search(cls, query: str, limit: int = 1) -> Iterator["Model"]:
        """Perform a Spotify search given a `query`

        Args:
            query (str): the search `query`
            limit (int, optional): the number of items to return. Defaults to 1.

        Raises:
            cls.NotFound: In case no results corresponding to the provided query
            are found

        Returns:
            Iterator["Model"]: the `Model` instances corresponding to the query
        """
        view_name = cls.view_name()

        results = cls.context.search(view_name, query, limit=limit)

        if not results:
            raise cls.NotFound

        return map(lambda kwargs: cls(**kwargs), results)

    @classmethod
    def validate_url(cls, method: Callable[..., Any]) -> Callable[..., Any]:
        """A decorator that validates the supplied `URL`
        before executing the decorated method

        Args:
            method (Callable[..., Any]): the method being decorated

        Raises:
            cls.InvalidURL: in case the supplied url is invalid

        Returns:
            Callable[..., Any]: the  decorated method
        """

        @wraps(method)
        def wrapper(cls, url, *args, **kwargs):
            view_name = cls.view_name()
            pattern = f"https://open.spotify.com/{view_name}"

            try:
                assert match(pattern, url) is not None
            except AssertionError:
                raise cls.InvalidURL from None

            return method(cls, url, *args, **kwargs)

        return wrapper

    @classmethod
    def http_safeguard(cls, method: Callable[..., Any]) -> Callable[..., Any]:
        """A decorator that converts http exceptions to `Model` level
        exceptions

        Args:
            method (Callable[..., Any]): the method being decorated

        Raises:
            cls.NotFound: in case a `Spotipy` call returns an HTTP status of 404
            cls.InvalidURL: in case a `Spotipy` call returns an HTTP status of 400
            cls.UnexpectedError: in case a `Spotipy` call returns any other HTTP status
            indicating an error

        Returns:
            Callable[..., Any]: the decorated method
        """

        @wraps(method)
        def wrapper(cls, *args, **kwargs):
            try:
                return method(cls, *args, **kwargs)
            except SpotifyException as e:
                if e.http_status == 404:
                    raise cls.NotFound from None
                elif e.http_status == 400:
                    raise cls.InvalidURL from None
                else:
                    raise cls.UnexpectedError from None

        return wrapper



from dotify.model import Model
from dotify.models.album import Album
from dotify.models.artist import Artist
from dotify.models.image import Image
from dotify.models.playlist import Playlist
from dotify.models.track import Track
from dotify.models.user import User

__all__ = [
    "Model",
    "Album",
    "Artist",
    "Image",
    "Playlist",
    "Track",
    "User",
]



import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List

from moviepy.editor import AudioFileClip
from mutagen.easyid3 import EasyID3
from pytube import YouTube
from pytube.streams import Stream
from youtubesearchpython import VideosSearch

from dotify.model import Model, logger

logger = logging.getLogger(f"{logger.name}.{__name__}")

EasyID3.RegisterTextKey("albumcover", "APIC")

if TYPE_CHECKING is True:
    from dotify.models.artist import Artist


class Track(Model):
    """ """

    class Json:
        """ """

        dependencies = [
            "dotify.models.Album",
            "dotify.models.Artist",
            "dotify.models.Image",
        ]

    def __str__(self) -> str:
        return f"{self.artist} - {self.name}"

    @property
    def url(self) -> str:
        """ """
        return self.external_urls.spotify

    @property
    def artist(self) -> "Artist":
        """ """
        return self.artists[0]

    @property
    def genres(self) -> List[Any]:
        """ """
        genres = []
        for item in [self.album, self.artist]:
            if hasattr(item, "genres"):
                genres.append(item.genres)

        return genres

    @property
    def genre(self) -> None:
        """ """
        return self.genres[0] if self.genres else None

    def streams(self, limit=1):
        """"""
        results = VideosSearch(str(self), limit=limit).result()["result"]

        for result in results:
            yield YouTube(result["link"]).streams.get_audio_only()

    @property
    def stream(self) -> Stream:
        """ """
        return next(self.streams(limit=1))

    @property
    def id3_tags(self) -> Dict[str, Any]:
        """ """
        optional: Dict[str, Any] = {}
        if self.genre is not None:
            optional["genre"] = self.genre

        return {
            **optional,
            "title": self.name,
            "titlesort": self.name,
            "tracknumber": str(self.track_number),
            "artist": [artist.name for artist in self.artists],
            "album": self.album.name,
            "albumartist": [artist.name for artist in self.album.artists],
            "date": self.album.release_date,
            "originaldate": self.album.release_date,
            "albumcover": self.album.cover,
        }

    def as_mp4(self, mp4_path: Path, skip_existing: bool = False) -> Path:
        """"""
        mp4_path = Path(mp4_path)

        return Path(
            self.stream.download(
                output_path=mp4_path.parent,
                filename=mp4_path.stem,
                skip_existing=skip_existing,
            )
        )

    def as_mp3(
        self, mp3_path: Path, skip_existing: bool = False, logger: None = None
    ) -> Path:
        """"""
        # FIXME: genres
        # FIXME: progress bar and logging both for moviepy and pytube

        mp3_path = Path(mp3_path)

        mp4_path = self.as_mp4(mp3_path, skip_existing=skip_existing)

        audio_file_clip = AudioFileClip(str(mp4_path))
        audio_file_clip.write_audiofile(str(mp3_path), logger=logger)

        mp4_path.unlink()

        easy_id3 = EasyID3(mp3_path)

        easy_id3.update(self.id3_tags)

        easy_id3.save(v2_version=3)

        return mp3_path

    def download(
        self, mp3_path: Path, skip_existing: bool = False, logger: None = None
    ) -> Path:
        """"""
        return self.as_mp3(mp3_path, skip_existing=skip_existing, logger=logger)

    @classmethod
    @Model.validate_url
    @Model.http_safeguard
    def from_url(cls, url: str) -> "Track":
        """"""
        return cls(**cls.context.track(url))
